from armcrop.detect import Detector, YOLODetector
from networkx.utils.union_find import UnionFind
import SimpleITK as sitk
import numpy as np
from collections import defaultdict


class UniqueObjectProcessor:
    def __init__(self, detector: YOLODetector, volume: sitk.Image):
        self.detector = detector
        self.volume_z_spacing = volume.GetSpacing()[2]

    def objects(
        self,
        grouping_iou,
        grouping_interval,
        grouping_min_depth,
    ):

        detections = self.detector.results
        if detections == {}:
            raise ValueError("No detections found. Please run the detector first.")
        box_groups = self._groups(
            detections, grouping_iou, grouping_interval, self.volume_z_spacing, grouping_min_depth
        )
        return box_groups

    def _groups(self, detections, iou, interval, spacing_z, discard_zlength):
        """
        Group detected objects based on their spatial proximity.
        """
        # convert the interval in mm to the number of slices
        interval = int(interval / spacing_z)
        discard_length = int(discard_zlength / spacing_z)
        # loop over the classes in the detections
        groups_xyz4 = defaultdict(list)
        for class_idx in detections:
            # detections are in the format (z, x, y, w, h, rotation, confidence, class_id)
            # ensure array is ordered by z
            if detections[class_idx] is None:
                continue
            # print(f"Processing class {class_idx} with {len(detections[class_idx])} detections")
            sorted_detections = sorted(detections[class_idx], key=lambda z: z[0])
            zs, xywhrs, _ = np.split(sorted_detections, [1, 6], axis=1)

            idxs = np.arange(zs.shape[0])
            matches = {}
            xyz4s = []
            for i in idxs:
                # get the z, xywhr for the current detection
                z = zs[i]
                xywhr = xywhrs[i].reshape(1, -1)
                # convert to xyz4 format
                xyxyxyxy = self._xywhr2xyxyxyxy(xywhr, round=True, pad=0)
                # add in z coordinate
                xyzxyzxyzxyz = np.concatenate([xyxyxyxy, np.full((1, 4, 1), z)], axis=-1)
                xyz4s.append(xyzxyzxyzxyz)

                # find nearby zs within the interval
                z_nearby_mask = ((zs >= z - interval) & (zs <= z + interval)).flatten()
                nearby_indices = np.where(z_nearby_mask)[
                    0
                ]  # Get original indices where mask is True
                xywhr_nearby = xywhrs[z_nearby_mask]

                # calculate IOU with nearby detections
                ious = np.triu(self.detector._batch_probiou(xywhr, xywhr_nearby), 1)
                pick = (ious > iou).flatten()

                # Store the original indices that match
                if len(pick) > 0 and np.any(pick):
                    matches[i] = nearby_indices[pick]
                else:
                    matches[i] = np.array([], dtype=int)

            # Concatenate all xyz4s
            xyz4s = np.concatenate(xyz4s, axis=0)

            # find overlapping unions
            ds = UnionFind()
            # Initialize all nodes in the UnionFind
            for i in idxs:
                ds[i]  # This ensures all indices are present in the UnionFind

            # Now create the unions based on matches
            for i, match_indices in matches.items():
                for j in match_indices:
                    if i != j:  # Don't union a node with itself
                        ds.union(i, j)

            # Get the sets from the UnionFind
            ds_sets = sorted([sorted(s) for s in ds.to_sets() if len(s) > 0], key=len, reverse=True)

            for group in ds_sets:
                if len(group) < discard_length:
                    continue
                groups_xyz4[class_idx].append(xyz4s[group, :].astype(int))

        return groups_xyz4

    def _xywhr2xyxyxyxy(self, x, round=True, pad=0):
        """
        Convert batched Oriented Bounding Boxes (OBB) from [xywh, rotation] to [xy1, xy2, xy3, xy4].
        Where xy1 is the top-left corner, xy2 is the top-right corner, xy3 is the bottom-right corner, and xy4 is the bottom-left corner.

        Args:
            x (numpy.ndarray): Boxes in [cx, cy, w, h, rotation] format of shape (n, 5) or (b, n, 5).

        Returns:
            (numpy.ndarray): Converted corner points of shape (n, 4, 2) or (b, n, 4, 2).
        """

        def round2pminf_add_remainder(x, r):
            return np.copysign(np.ceil(np.abs(x) + r), x)

        def regularize_rboxes(rboxes):
            """
            Regularize rotated boxes in range [0, pi/2].

            Args:
                rboxes (numpy.ndarray): Input boxes of shape(N, 5) in xywhr format.

            Returns:
                (numpy.ndarray): The regularized boxes.
            """
            x, y, w, h, t = np.split(rboxes, 5, axis=-1)
            w_ = np.where(w > h, w, h)
            h_ = np.where(w > h, h, w)
            t = np.where(w > h, t, t + np.pi / 2) % np.pi
            return np.concatenate([x, y, w_, h_, t], axis=-1)  # regularized boxes

        # Regularize the input boxes first
        rboxes = regularize_rboxes(x)

        ctr = rboxes[..., :2]
        w, h, angle = (rboxes[..., i : i + 1] for i in range(2, 5))

        # add padding to the width and height
        w += pad
        h += pad

        cos_value = np.cos(angle)
        sin_value = np.sin(angle)

        vec1 = np.concatenate([w / 2 * cos_value, w / 2 * sin_value], axis=-1)
        vec2 = np.concatenate([-h / 2 * sin_value, h / 2 * cos_value], axis=-1)

        # round to the nearest pixel
        if round:
            rem = np.remainder(ctr, np.floor(ctr))
            ctr = np.floor(ctr)
            pt1 = ctr + round2pminf_add_remainder(vec1 + vec2, rem)
            pt2 = ctr + round2pminf_add_remainder(vec1 - vec2, rem)
            pt3 = ctr - round2pminf_add_remainder(vec1 - vec2, rem)
            pt4 = ctr - round2pminf_add_remainder(vec1 + vec2, rem)
        else:
            pt1 = ctr + vec1 + vec2
            pt2 = ctr + vec1 - vec2
            pt3 = ctr - vec1 - vec2
            pt4 = ctr - vec1 + vec2

        return np.stack([pt1, pt2, pt3, pt4], axis=-2)


if __name__ == "__main__":
    # load a volume
    volume = sitk.ReadImage("/mnt/slowdata/ct/cadaveric-full-arm/1606011L/1606011L.nrrd")
    # Example usage
    detector = YOLODetector(model_type="yolo11-obb")

    detector.predict(
        volume,
        confidence_threshold=0.5,
        iou_threshold=0.5,
    )
    groups_detector = UniqueObjectProcessor(detector, volume)

    groups_detector.objects(
        grouping_iou=0.2,  # Example IOU threshold for grouping
        grouping_interval=50,  # Example interval
        grouping_min_depth=20,  # Example minimum depth for grouping
    )



from armcrop.group import UniqueObjectProcessor
from armcrop.detect import YOLODetector
import abc
import SimpleITK as sitk
from typing import List
import numpy as np
from math import ceil


class CropBase(abc.ABC):
    _detector_cache = {}  # Class-level cache for YOLODetector instances
    _MAX_CACHE_SIZE = 1  # Maximum number of items in the cache

    def __init__(
        self,
        volume: sitk.Image,
        detection_confidence: float = 0.5,
        detection_iou: float = 0.5,
    ):
        """
        Initialize the cropper with a volume.

        Args:
            volume (sitk.Image): The volume to be cropped. Must be a SimpleITK image.
            detection_confidence (float): Confidence threshold for the YOLO detector.
            detection_iou (float): IOU threshold for the YOLO detector.
        """

        self._volume = volume

        # Using id(volume) for the cache key. This assumes that for the "same" volume,
        # the exact same sitk.Image object instance is used.
        # If not, a more robust key based on volume properties would be needed.
        cache_key = (id(self._volume), float(detection_confidence), float(detection_iou))

        if cache_key in CropBase._detector_cache:
            self._det = CropBase._detector_cache[cache_key]
            # print(f"Reusing cached detector for volume ID {id(self.volume)} and params ({detection_confidence}, {detection_iou})") # Optional: for debugging
        else:
            # If the cache is full (i.e., contains _MAX_CACHE_SIZE items)
            # and we are about to add a new, different item, clear the cache first.
            if len(CropBase._detector_cache) >= CropBase._MAX_CACHE_SIZE:
                # print(f"Cache limit ({CropBase._MAX_CACHE_SIZE}) reached. Clearing cache before adding new item.") # Optional: for debugging
                CropBase._detector_cache.clear()

            # print(f"Creating new detector for volume ID {id(self.volume)} and params ({detection_confidence}, {detection_iou})") # Optional: for debugging
            self._det = YOLODetector()
            self._det.predict(self._volume, detection_confidence, detection_iou)
            CropBase._detector_cache[cache_key] = self._det

        self._uop = UniqueObjectProcessor(self._det, self._volume)
        self._lblmap = {
            "clavicle": 0,
            "scapula": 1,
            "humerus": 2,
            "radius-ulna": 3,
            "hand": 4,
        }

    def _check_bone_request(self, requested_bone):
        if requested_bone not in self._lblmap.keys():
            raise ValueError(
                f"Requested bone '{requested_bone}' is not valid. Options are: {list(self._lblmap.keys())}"
            )


class Crop(CropBase):
    def __init__(self, volume, detection_confidence=0.5, detection_iou=0.5):
        super().__init__(volume, detection_confidence, detection_iou)
        self.interpolator = sitk.sitkBSpline

    def _boolean_volume(self, xyz4):
        """Convert the xyz boxes to a boolean mask of all vertices, handles out of bounds errors."""

        # reshape from (n, 4, 3) to (n, 3)
        xyz = xyz4.reshape(-1, 3)

        # clip along x,y,z axes to ensure we don't go out of bounds
        xyz[:, 0] = np.clip(xyz[:, 0], 0, self._volume.GetSize()[0] - 1)
        xyz[:, 1] = np.clip(xyz[:, 1], 0, self._volume.GetSize()[1] - 1)
        xyz[:, 2] = np.clip(xyz[:, 2], 0, self._volume.GetSize()[2] - 1)

        # we are flipping the axes to match the sitk coordinate system
        zyx_bool = np.zeros(np.flip(self._volume.GetSize()))
        zyx_bool[xyz[:, 2], xyz[:, 1], xyz[:, 0]] = 1
        zyx_bool = sitk.GetImageFromArray(zyx_bool)
        zyx_bool.CopyInformation(self._volume)
        zyx_bool = sitk.Cast(zyx_bool, sitk.sitkUInt8)
        return zyx_bool

    def _filters(self, box_groups):
        filters = []
        for box_group in box_groups:
            if len(box_group) == 0:
                continue
            # get boolean mask of box vertices
            zyx_bool = self._boolean_volume(box_group)
            bb_filter = sitk.LabelShapeStatisticsImageFilter()
            bb_filter.Execute(zyx_bool)
            filters.append(bb_filter)

        return filters

    def _process(self, box_groups, spacing):
        cropped_imgs = []
        # get boolean mask of box vertices
        bb_filters = self._filters(box_groups)
        for bb_filter in bb_filters:
            # crop the image using the bounding box
            boundingbox = bb_filter.GetBoundingBox(1)
            roi_filter = sitk.RegionOfInterestImageFilter()
            roi_filter.SetRegionOfInterest(boundingbox)
            crop_img = roi_filter.Execute(self._volume)
            # resample the cropped image to the desired spacing
            resampler = sitk.ResampleImageFilter()
            resampler.SetOutputDirection(crop_img.GetDirection())
            resampler.SetOutputOrigin(crop_img.GetOrigin())
            resampler.SetOutputSpacing(spacing)
            resampler.SetSize(
                [
                    int(ceil(crop_img.GetSize()[i] * crop_img.GetSpacing()[i] / spacing[i]))
                    for i in range(3)
                ]
            )
            resampler.SetInterpolator(self.interpolator)
            resampler.SetDefaultPixelValue(-1023)
            crop_img = resampler.Execute(crop_img)

            cropped_imgs.append(crop_img)
        return cropped_imgs

    def process(
        self,
        bone: str,
        grouping_iou: float = 0.5,
        grouping_interval: int = 40,
        grouping_min_depth: int = 50,
        spacing: tuple = (0.5, 0.5, 0.5),
    ) -> List[sitk.Image]:
        """
        Process the volume for the requested bone.

        Args:
            bone: The bone to be cropped. Options are: clavicle, scapula, humerus, radius-ulna, hand.
            grouping_iou: IOU threshold for grouping.
            grouping_interval: Interval for grouping.
            grouping_min_depth: Minimum depth in mm for grouping detection boxes together.
            obb_spacing: Spacing for the resampling.

        Returns:
            Aligned images of the oriented bounding boxes.
        """
        self._check_bone_request(bone)
        box_groups = self._uop.objects(
            grouping_iou,
            grouping_interval,
            grouping_min_depth,
        )
        box_groups = box_groups[self._lblmap[bone]]
        return self._process(box_groups, spacing)


class CropOriented(Crop):
    def __init__(self, volume, detection_confidence=0.5, detection_iou=0.5):
        super().__init__(volume, detection_confidence, detection_iou)

    def _filters(self, box_groups):
        filters = []
        for box_group in box_groups:
            if len(box_group) == 0:
                continue
            # get boolean mask of box vertices
            zyx_bool = self._boolean_volume(box_group)

            obb_filter = sitk.LabelShapeStatisticsImageFilter()
            obb_filter.ComputeOrientedBoundingBoxOn()
            obb_filter.Execute(zyx_bool)
            filters.append(obb_filter)

        return filters

    def _process(self, box_groups, spacing):
        aligned_imgs = []
        # get boolean mask of box vertices
        obb_filters = self._filters(box_groups)
        for obb_filter in obb_filters:
            aligned_img_dir = (
                np.array(obb_filter.GetOrientedBoundingBoxDirection(1))
                .reshape(3, 3)
                .T.flatten()
                .tolist()
            )
            # boot up the resampler
            resampler = sitk.ResampleImageFilter()
            # calculate the size of the aligned image
            aligned_img_size = [
                int(ceil(obb_filter.GetOrientedBoundingBoxSize(1)[i] / spacing[i]))
                for i in range(3)
            ]
            # set origin for the aligned image
            obb_origin = obb_filter.GetOrientedBoundingBoxOrigin(1)

            resampler.SetOutputDirection(aligned_img_dir)
            resampler.SetOutputOrigin(obb_origin)
            resampler.SetOutputSpacing(spacing)
            resampler.SetSize(aligned_img_size)
            resampler.SetInterpolator(self.interpolator)
            resampler.SetDefaultPixelValue(-1023)

            # get the aligned image, and its array
            aligned_img = resampler.Execute(self._volume)

            aligned_imgs.append(aligned_img)
        return aligned_imgs


class Centroids(CropBase):
    def __init__(self, volume, detection_confidence=0.5, detection_iou=0.5):
        super().__init__(volume, detection_confidence, detection_iou)

    def _process(self, box_groups) -> List[np.ndarray]:
        """Calculate the centroids of the detected boxes."""
        centroids = []
        for box_group in box_groups:
            if len(box_group) == 0:
                continue
            # Calculate the centroid of each box
            centroid_idx = np.mean(box_group, axis=1).astype(int)
            centroids_obj = []
            for ci in centroid_idx:
                ci_mm = self._volume.TransformIndexToPhysicalPoint(ci.tolist())
                # ci2 = self.volume.TransformPhysicalPointToIndex(ci_mm)
                # print(ci, ci_mm, ci2)
                centroids_obj.append(ci_mm)
            centroids.append(np.array(centroids_obj))
        return centroids

    def process(
        self,
        bone: str,
        grouping_iou: float = 0.5,
        grouping_interval: int = 40,
        grouping_min_depth: int = 50,
    ) -> List[np.ndarray]:
        """
        Process the volume to get centroids of detected boxes.

        Args:
            bone: The bone to be processed. Options are: clavicle, scapula, humerus, radius-ulna, hand.
            grouping_iou: IOU threshold for grouping.
            grouping_interval: Interval for grouping.
            grouping_min_depth: Minimum depth in mm for grouping detection boxes together.

        Returns:
            List of centroids for each detected box.
        """
        self._check_bone_request(bone)
        box_groups = self._uop.objects(
            grouping_iou,
            grouping_interval,
            grouping_min_depth,
        )
        box_groups = box_groups[self._lblmap[bone]]
        return self._process(box_groups)


if __name__ == "__main__":
    volume = sitk.ReadImage("/mnt/slowdata/ct/cadaveric-full-arm/09-12052L/09-12052L.nrrd")
    # Example: Create another reference to the same volume object
    # volume_ref2 = volume
    # If you were to load it again:
    # volume2 = sitk.ReadImage("/mnt/slowdata/ct/cadaveric-full-arm/09-12052L/09-12052L.nrrd")
    # Then id(volume) != id(volume2), and caching based on id() would not work across these two.

    OBB = True
    CENTROID = True  # Run both to test caching
    CROP = True  # Run all three
    OTHER_SETTINGS = False  # Test with different detection parameters
    OTHER_VOLUME = True  # Test with a different volume

    det_conf = 0.2
    det_iou = 0.2

    print(f"Initial cache size: {len(CropBase._detector_cache)}")

    if OBB:
        print("Processing OBB...")
        cropper_obb = CropOriented(
            volume,  # Pass the same volume object
            detection_confidence=det_conf,
            detection_iou=det_iou,
        )
        output_obb = cropper_obb.process(
            bone="scapula",
            grouping_iou=0.2,
            grouping_interval=50,
            grouping_min_depth=20,
            spacing=(1.0, 1.0, 1.0),
        )
        for i, img in enumerate(output_obb):
            sitk.WriteImage(img, f"aligned_{i}.nrrd")
        print(f"Saved {len(output_obb)} OBB images.")
        print(f"Cache size after OBB: {len(CropBase._detector_cache)}")

    if CENTROID:
        print("Processing Centroids...")
        # Using the same volume object and detection parameters
        centroider = Centroids(volume, detection_confidence=det_conf, detection_iou=det_iou)
        output_centroids = centroider.process(
            "humerus",
            grouping_iou=0.2,
            grouping_interval=50,
            grouping_min_depth=20,
        )
        for i, centroid_val in enumerate(output_centroids):
            print(f"Centroid {i}: {centroid_val}")
        print(f"Found {len(output_centroids)} centroids.")
        print(f"Cache size after Centroids: {len(CropBase._detector_cache)}")

    if CROP:
        print("Processing Axis-Aligned Crop...")
        # Using the same volume object and detection parameters
        cropper_axis_aligned = Crop(
            volume,
            detection_confidence=det_conf,
            detection_iou=det_iou,
        )
        output_crop = cropper_axis_aligned.process(
            bone="humerus",
            grouping_iou=0.2,
            grouping_interval=50,
            grouping_min_depth=20,
            spacing=(1.0, 1.0, 1.0),
        )
        for i, img in enumerate(output_crop):
            sitk.WriteImage(img, f"cropped_{i}.nrrd")
        print(f"Saved {len(output_crop)} cropped images.")
        print(f"Cache size after Crop: {len(CropBase._detector_cache)}")

    # Example with different detection parameters - this would create a new cache entry
    if OTHER_SETTINGS:
        print("Processing OBB with different detection parameters...")
        cropper_obb_diff_params = CropOriented(
            volume,  # Same volume object
            detection_confidence=0.7,  # Different confidence
            detection_iou=det_iou,
        )
        # ... process and use cropper_obb_diff_params ...
        print(f"Cache size after OBB with different params: {len(CropBase._detector_cache)}")

        print("Processing Centroids with same different parameters...")
        centroider_diff_params = Centroids(
            volume,  # Same volume object
            detection_confidence=0.7,  # Different confidence
            detection_iou=det_iou,
        )
        output_centroids_diff = centroider_diff_params.process(
            "humerus",
            grouping_iou=0.2,
            grouping_interval=50,
            grouping_min_depth=20,
        )
        for i, centroid_val in enumerate(output_centroids_diff):
            print(f"Centroid {i}: {centroid_val}")
        print(f"Found {len(output_centroids_diff)} centroids with different params.")
        print(f"Cache size after Centroids with different params: {len(CropBase._detector_cache)}")

    if OTHER_VOLUME:
        print("Processing OBB with a different volume...")
        # Load a different volume
        volume2 = sitk.ReadImage("/mnt/slowdata/ct/cadaveric-full-arm/1606011L/1606011L.nrrd")
        cropper_obb_diff_volume = CropOriented(
            volume2,  # Different volume object
            detection_confidence=det_conf,
            detection_iou=det_iou,
        )
        output_obb_diff_volume = cropper_obb_diff_volume.process(
            bone="scapula",
            grouping_iou=0.2,
            grouping_interval=50,
            grouping_min_depth=20,
            spacing=(1.0, 1.0, 1.0),
        )
        for i, img in enumerate(output_obb_diff_volume):
            sitk.WriteImage(img, f"aligned_diff_{i}.nrrd")
        print(f"Saved {len(output_obb_diff_volume)} OBB images for the new volume.")
        print(f"Cache size after OBB with different volume: {len(CropBase._detector_cache)}")



import onnxruntime as rt
import numpy as np
import pathlib
import huggingface_hub
from typing import Dict


class ModelManager:
    """
    Singleton class that loads and caches ML models for bone detection
    """

    _instances: Dict[str, "ModelManager"] = {}
    _available_models = ["yolo11-obb", "yolo9-bb"]  # corresponds to name in _get_model_path

    @classmethod
    def get_instance(cls, model_type):
        if model_type not in cls._instances:
            cls._instances[model_type] = cls(model_type)
        return cls._instances[model_type]

    def __init__(self, model_type):
        self.model = None
        self.model_type = model_type
        self.img_size = (640, 640)

    def load_model(self):
        """Load the model if it hasn't been loaded already"""
        if self.model is not None:
            return self.model

        model_path = self._get_model_path()

        # load model
        with open(model_path, "rb") as file:
            try:
                import torch

                providers = [
                    (
                        "CUDAExecutionProvider",
                        {
                            "device_id": torch.cuda.current_device(),
                            "user_compute_stream": str(torch.cuda.current_stream().cuda_stream),
                        },
                    )
                ]
            except:
                print("Using CPUExecutionProvider")
                providers = ["CPUExecutionProvider"]
            self.model = rt.InferenceSession(file.read(), providers=providers)

        # prime the model on a random image
        self.model.run(
            None,
            {"images": np.random.rand(1, 3, self.img_size[0], self.img_size[1]).astype(np.float32)},
        )

        return self.model

    def _get_model_path(self):
        """Get the appropriate model path based on model_type"""
        if self.model_type == self._available_models[0]:  # yolo11-obb
            return pathlib.Path(
                huggingface_hub.hf_hub_download(
                    repo_id="gregspangenberg/armcrop",
                    filename="upperarm_yolo11_obb_6.onnx",
                )
            )
        elif self.model_type == self._available_models[1]:  # yolo9-bb
            return pathlib.Path(
                huggingface_hub.hf_hub_download(
                    repo_id="gregspangenberg/armcrop",
                    filename="yolov9c_upperlimb.onnx",
                )
            )
        else:
            raise ValueError(
                f"Unsupported model type: {self.model_type}\n Available types: {self._available_models}"
            )



import pathlib
import numpy as np
from typing import Tuple, List, Dict
import onnxruntime as rt
import SimpleITK as sitk

from copy import deepcopy
from armcrop.model_manager import ModelManager
import abc

# disable logging
rt.set_default_logger_severity(3)


class Detector(abc.ABC):
    """
    Base class for cropping volumes using a machine learning model.
    """

    def __init__(self, img_size=(640, 640), model_type="default"):
        """
        Initialize the BaseCrop class.

        Args:
            img_size: Image size required for the ML model. Defaults to (640, 640).
        """
        self.img_size = img_size
        self.model = None
        self.model_type = model_type
        self.results = {}

    def format_volume(self, volume: sitk.Image, img_size=(640, 640)) -> sitk.Image:
        """
        Load a volume and preprocess it for inference

        Args:
            volume: Path to the input volume for inference, or the sitk.Image volume
            img_size: Image size required for the ML model. Defaults to (640, 640).

        Returns:
            vol: The original volume
            vol_t: The preprocessed volume
        """
        vol_t = deepcopy(volume)
        vol_t = sitk.Cast(vol_t, sitk.sitkFloat32)
        vol_t = sitk.Clamp(vol_t, sitk.sitkFloat32, -1024, 3000)
        vol_t = sitk.RescaleIntensity(vol_t, 0, 1)

        new_size = [img_size[0], img_size[1], vol_t.GetDepth()]

        reference_image = sitk.Image(new_size, vol_t.GetPixelIDValue())
        reference_image.SetOrigin(vol_t.GetOrigin())
        reference_image.SetDirection(vol_t.GetDirection())
        reference_image.SetSpacing(
            [sz * spc / nsz for nsz, sz, spc in zip(new_size, vol_t.GetSize(), vol_t.GetSpacing())]
        )

        vol_t = sitk.Resample(vol_t, reference_image)

        return vol_t

    def load_model(self) -> rt.InferenceSession:
        """Load the ML model for inference.

        Returns:
            rt.InferenceSession: The loaded ML model.
        """
        # Use the model manager to get the cached model
        model_manager = ModelManager.get_instance(self.model_type)
        model = model_manager.load_model()

        return model

    def _preprocess_array(self, arr) -> np.ndarray:
        arr = np.expand_dims(arr, axis=0)
        arr = np.expand_dims(arr, axis=0)
        arr = np.repeat(arr, 3, axis=1)
        arr = arr.astype(np.float32)

        return arr

    def _class_dict_construct(self, data) -> Dict:
        """
        Creates the dictionary with the data for each class

        Args:
            data: The data from the ML model

        Returns:
            class_data: The dictionary with the ml data broken down by class
        """
        # break up by class
        sorted_cls_idx = np.argsort(data[:, -1])
        sorted_data = data[sorted_cls_idx]
        unique_cls, unique_cls_idx = np.unique(sorted_data[:, -1], return_index=True)
        split_data = np.split(sorted_data, unique_cls_idx[1:])

        # Create a dictionary with keys from 0 to 4, defaulting to None
        class_data = {cls: None for cls in range(5)}

        # Update the dictionary with the actual data
        for cls, cls_data in zip(unique_cls, split_data):
            # sort by z
            cls_data = cls_data[np.argsort(cls_data[:, -1])]
            class_data[int(cls)] = cls_data

        return class_data

    def predict(
        self, volume: sitk.Image, confidence_threshold, iou_threshold
    ) -> Dict[int, np.ndarray]:
        """
        Run inference on the volume and return the cropped volume.

        Args:
            volume: The input volume for inference.
            confidence_threshold: Confidence threshold for predictions.
            iou_threshold: IoU threshold for non-max suppression.

        Returns:
            A dictionary with class indices as keys and numpy arrays of predictions as values.
        """
        raise NotImplementedError("This method should be implemented by subclasses.")


class YOLODetector(Detector):
    """
    Detector for cropping volumes using an Oriented Bounding Box (OBB) model.
    """

    def __init__(self, img_size=(640, 640), model_type="yolo11-obb"):
        super().__init__(img_size, model_type)

    def nms_rotated_batch(self, prediction, conf_thres, iou_thres) -> List[np.ndarray]:
        """
        Perform non-maximum suppression (NMS) on a set of boxes, with support for masks and multiple labels per box.

        Args:
            prediction (np.ndarray): A tensor of shape (batch_size, num_classes + 4 + num_masks, num_boxes)
                containing the predicted boxes, classes, and masks. The tensor should be in the format
                output by a model, such as YOLO.
            conf_thres (float): The confidence threshold below which boxes will be filtered out.
                Valid values are between 0.0 and 1.0.
            iou_thres (float): The IoU threshold below which boxes will be filtered out during NMS.
                Valid values are between 0.0 and 1.0.
            nc (int, optional): The number of classes output by the model. Any indices after this will be considered masks.
        Returns:
            List[np.ndarray]: A list of numpy arrays, each containing the filtered boxes for each image in the batch.
             The columns are x,y,w,h, rotation, confidence,class_id
        """

        # Checks
        assert (
            0 <= conf_thres <= 1
        ), f"Invalid Confidence threshold {conf_thres}, valid values are between 0.0 and 1.0"
        assert 0 <= iou_thres <= 1, f"Invalid IoU {iou_thres}, valid values are between 0.0 and 1.0"

        bs = prediction.shape[0]  # batch size (BCN,. Defaults to 1,84,6300)
        nc = 5  # number of classes
        nm = prediction.shape[1] - nc - 4  # number of masks
        mi = 4 + nc  # mask start index
        xc = np.max(prediction[:, 4:mi], axis=1) > conf_thres
        prediction = prediction.transpose((0, 2, 1))  # shape(1,84,6300) to shape(1,6300,84)

        output = [np.zeros((0, 6 + nm))] * bs

        for xi, x in enumerate(prediction):  # image index, image inference
            x = x[xc[xi]]  # confidence

            # If none remain process next image
            if not x.shape[0]:
                continue

            # Detections matrix nx6 (xywh, conf, cls, rot)
            box = x[:, :4]  # xywh
            cls = x[:, 4:mi]  # classes
            rot = x[:, mi : mi + 1]  # rotation
            if rot.shape[1] == 0:  # if rotation is not present, add a column of zeros
                rot = np.zeros((box.shape[0], 1), dtype=np.float32)
            conf = np.max(cls, axis=1, keepdims=True)
            class_id = np.argmax(cls, axis=1, keepdims=True)

            x = np.concatenate((box, rot, conf, class_id), 1)[conf.squeeze(1) > conf_thres]

            if not x.shape[0]:  # no boxes
                continue

            # Batched NMS
            boxes = x[:, :5]  # xywhr
            scores = x[:, 5]  # scores

            i = self._nms_rotated(boxes, scores, iou_thres)
            output[xi] = x[i]

        return output

    def _nms_rotated(self, boxes, scores, threshold) -> np.ndarray:
        """
        NMS for oriented bounding boxes using probiou and fast-nms.

        Args:
            boxes (np.ndarray): Rotated bounding boxes, shape (N, 5), format xywhr.
            scores (np.ndarray): Confidence scores, shape (N,).
            threshold (float, optional): IoU threshold

        Returns:
            (np.ndarray): Indices of boxes to keep after NMS.
        """

        if len(boxes) == 0:
            return np.empty((0,), dtype=np.int8)

        # Sort boxes by confidence scores in descending order
        sorted_idx = np.argsort(scores, axis=-1)[::-1]
        boxes = boxes[sorted_idx]

        ious = np.triu(self._batch_probiou(boxes, boxes), 1)
        # Filter boxes based on IOU threshold to remove overlapping detections

        pick = np.nonzero(ious.max(axis=0) < threshold)[0]

        return sorted_idx[pick]

    def _batch_probiou(self, obb1, obb2, eps=1e-7) -> np.ndarray:
        """
        Calculate the prob IoU between oriented bounding boxes, https://arxiv.org/pdf/2106.06072v1.pdf.

        Args:
            obb1 (np.ndarray): A tensor of shape (N, 5) representing ground truth obbs, with xywhr format.
            obb2 (np.ndarray): A tensor of shape (M, 5) representing predicted obbs, with xywhr format.
            eps (float, optional): A small value to avoid division by zero. Defaults to 1e-7.

        Returns:
            (np.ndarray): A tensor of shape (N, M) representing obb similarities.
        """

        # Split coordinates into separate x,y components for both sets of boxes
        x1, y1 = np.split(obb1[..., :2], 2, axis=-1)
        x2, y2 = (x.squeeze(-1)[None] for x in np.split(obb2[..., :2], 2, axis=-1))

        # Calculate covariance matrices for both sets of boxes
        a1, b1, c1 = self._get_covariance_matrix(obb1)
        a2, b2, c2 = (x.squeeze(-1)[None] for x in self._get_covariance_matrix(obb2))

        # Calculate first term of Bhattacharyya distance using quadratic forms
        t1 = (
            ((a1 + a2) * np.power((y1 - y2), 2) + (b1 + b2) * np.power((x1 - x2), 2))
            / ((a1 + a2) * (b1 + b2) - np.power((c1 + c2), 2) + eps)
        ) * 0.25

        # Calculate second term (cross term) of Bhattacharyya distance
        t2 = (
            ((c1 + c2) * (x2 - x1) * (y1 - y2))
            / ((a1 + a2) * (b1 + b2) - np.power((c1 + c2), 2) + eps)
        ) * 0.5

        # Calculate third term (determinant term) of Bhattacharyya distance
        t3 = (
            np.log(
                ((a1 + a2) * (b1 + b2) - np.power((c1 + c2), 2))
                / (
                    4
                    * np.sqrt(
                        (a1 * b1 - (np.power(c1, 2)).clip(0))
                        * (a2 * b2 - (np.power(c2, 2)).clip(0))
                    )
                    + eps
                )
                + eps
            )
            * 0.5
        )
        # Convert Bhattacharyya distance to Hellinger distance and then to IOU
        bd = np.clip(t1 + t2 + t3, eps, 100.0)
        hd = np.sqrt(1 - np.exp(-bd) + eps)

        return 1 - hd

    def _get_covariance_matrix(self, boxes) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Generating covariance matrix from obbs.

        Args:
            boxes (np.ndarray): A tensor of shape (N, 5) representing rotated bounding boxes, with xywhr format.

        Returns:
            (np.ndarray): Covariance matrices corresponding to original rotated bounding boxes.
        """
        # Gaussian bounding boxes, ignore the center points (the first two columns) because they are not needed here.
        gbbs = np.concatenate((np.power(boxes[:, 2:4], 2) / 12, boxes[:, 4:5]), axis=-1)
        a, b, c = np.split(gbbs, 3, axis=-1)
        cos = np.cos(c)
        sin = np.sin(c)
        cos2 = np.power(cos, 2)
        sin2 = np.power(sin, 2)
        return a * cos2 + b * sin2, a * sin2 + b * cos2, (a - b) * cos * sin

    def predict(
        self, volume: sitk.Image, confidence_threshold, iou_threshold
    ) -> Dict[int, np.ndarray]:
        """
        Run inference on the volume and return the cropped volume.

        Args:
            volume: The input volume for inference.
            confidence_threshold: Confidence threshold for predictions.
            iou_threshold: IoU threshold for non-max suppression.

        Returns:
            vol: The original volume.
            vol_t: The cropped volume based on the OBB model predictions.
        """
        model = self.load_model()
        original_size = (volume.GetWidth(), volume.GetHeight())
        vol_t = self.format_volume(volume, self.img_size)
        # get original width of the volume

        # Process each axial slice for prediction
        data = []
        for z in range(vol_t.GetDepth()):
            # Prepare the 2D slice for model inference by expanding dimensions and repeating channels
            arr = sitk.GetArrayFromImage(vol_t[:, :, z])
            arr = self._preprocess_array(arr)
            # run inference on the image
            output = model.run(None, {"images": arr})[0]

            # perform non-max supression on the output
            preds = self.nms_rotated_batch(output, confidence_threshold, iou_threshold)[0]

            # convert xywh out of 640 to original volume size
            preds[:, 0] = preds[:, 0] * original_size[0] / self.img_size[0]
            preds[:, 1] = preds[:, 1] * original_size[1] / self.img_size[1]
            preds[:, 2] = preds[:, 2] * original_size[0] / self.img_size[0]
            preds[:, 3] = preds[:, 3] * original_size[1] / self.img_size[1]

            # add the z coordinate to the predictions
            preds = np.c_[np.ones((preds.shape[0], 1)) * z, preds]
            data.extend(preds)

        # organize predictions by class
        # output is z, xywhr, confidence, class_id,
        cls_dict = self._class_dict_construct(np.array(data))
        self.results = cls_dict

        return self.results


if __name__ == "__main__":
    # Example usage
    detector = YOLODetector()
    volume = sitk.ReadImage("/mnt/slowdata/ct/cadaveric-full-arm/1606011L/1606011L.nrrd")
    conf_threshold = 0.5
    iou_threshold = 0.5

    predictions = detector.predict(volume, conf_threshold, iou_threshold)
    for cls, preds in predictions.items():
        print(f"Class {cls}: {preds.shape[0]} predictions")
        print(preds.shape)
        print(np.max(preds[:, 1]), np.min(preds[:, 1]))
